/*
    GL-117
    Copyright 2001, 2002 Thomas A. Drexl aka heptargon

    This file is part of GL-117.

    GL-117 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GL-117 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Jabytad; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef IS_MODEL_H
#define IS_MODEL_H

//#define USE_GLUT

#ifdef USE_GLUT
#include <GL/glut.h>
#else
#include <GL/glut.h>
//#include <GL/gl.h>
//#include <GL/glu.h>
#endif

// CColor stores color information
class CColor
{
  public:
  unsigned char c [4];
  CColor ();
  CColor (CColor *col);
  CColor (short cr, short cg, short cb);
  CColor (short cr, short cg, short cb, short ca);
  ~CColor ();
  void setColor (CColor *col);
  void setColor (short cr, short cg, short cb, short ca);
  void setColor (short cr, short cg, short cb);
  bool isEqual (CColor *col);
  void take (CColor *col);
};

// CTexture loads and stores a texture to memory
class CTexture
{
  public:
  unsigned char *data;
  char name [256];
  bool mipmap;
  int textureID;
  int width, height;
  float texlight;
  int quality;
  CTexture ();
  int loadFromTGA (char *fname, int quality, int alphatype, int mipmap);
  void getColor (CColor *c, int x, int y);
};

// CVector3 stores the components of a 3D vector
class CVector3 
{
  public:
  float x, y, z;
  CVector3 ();
  CVector3 (float x, float y, float z);
  CVector3 (CVector3 *v);
  void neg ();
  void add (CVector3 *v);
  void sub (CVector3 *v);
  void mult (CVector3 *v);
  float dotproduct (CVector3 *v);
  float length ();
  void norm ();
  bool isEqual (CVector3 *v);
  bool isEqual (CVector3 *v, float tol);
  void take (CVector3 *v);
};

// CVector2 stores the components of a 2D vector, for texture purpose
class CVector2 
{
  public:
  float x, y;
  void take (CVector2 *v);
};

// CVertex represents a vertex which may take inforamtion about color, location, normal, texture, number of surrounding triangles
class CVertex
{
  public:
  int triangles;
  CColor color;
  CVector3 vector;
  CVector3 normal;
  CVector2 tex;
  CVertex ();
  CVertex (CVertex *v);
  void addNormal (CVector3 *n);
  void addColor (CColor *c);
  void take (CVertex *v);
};

extern double pitab;
extern float sintab [360], costab [360];

// CRotation stores one (x,y,z)-rotation
class CRotation
{
  private:
  float rot [3] [3];
  public:
  short a, b, c;
  CRotation ();
  ~CRotation ();
  void setAngles (short a, short b, short c);
  void addAngles (short a, short b, short c);
  void calcRotation ();
  float rotateX (CVector3 *v);
  float rotateY (CVector3 *v);
  float rotateZ (CVector3 *v);
  float getsintab (int a);
  float getcostabntab (int a);
  void take (CRotation *r);
};

// CTriangle stores references to the vertices of the triangle/face
class CTriangle
{
  public:
  CVertex *v [3];
  void getNormal (CVector3 *n);
  void setVertices (CVertex *a, CVertex *b, CVertex *c);
};

// CQuad stores references to the vertices of the quad/square/face
class CQuad
{
  public:
  CVertex *v [4];
  void getNormal (CVector3 *n);
  void setVertices (CVertex *a, CVertex *b, CVertex *c, CVertex *d);
};

// CMaterial stores the name, filename, color, and texture of a material
class CMaterial
{
  public:
  char name [255];
  char filename [255];
  CColor color;
  CTexture *texture;
  float uTile;
  float vTile;
  float uOffset;
  float vOffset;
};

// CObject stores the material, vertices, and faces (triangles, quads) of an object
class CObject 
{
  public:
  int numVertices;
  int numTriangles;
  int numQuads;
  int numTexVertex;
  CMaterial *material;
  bool hasTexture;
  char name [255];
  CVertex *vertex;
  CTriangle *triangle;
  CQuad *quad;
  CObject ();
  ~CObject ();
  int addVertex (CVertex *w);
  void setColor (CColor *col);
};

// CModel stores the materials and objects of a model
class CModel 
{
  private:
  int rotcol;
  float light_ambient [4];
  float light_diffuse [4];
  float light_ambient2 [4];
  float light_diffuse2 [4];
  public:
  char name [20];
  int numObjects;
  int numMaterials;
  int list1, list2, list3;
  CMaterial *material [100];
  CObject *object [100];
  bool nolight;
  CModel ();
  void setName (char *name);
  void addMaterial (CMaterial *material);
  void addObject (CObject *object);
  ~CModel ();
  void setColor (CColor *col);
  void drawVertexNormals (CObject *cm, float zoom);
  int rotateColor (int n);
  void draw (CVector3 *tl, CVector3 *tl2, CRotation *rot, float zoom, float lum, int explode);
  void draw2 (CVector3 *tl, CVector3 *tl2, CRotation *rot, float zoom, int explode);
  void draw3 (CVector3 *tl, CVector3 *tl2, CRotation *rot, float zoom, int explode);
  void draw3 (CVector3 *tl, CVector3 *tl2, CRotation *rot, float zoom, float lum, int explode);
};

// CSphere represents a sphere according to the CModel data structure
class CSphere : public CModel
{
  private:
  int random (int n);
  public:
  int segments;
  float radius;
  float dx, dy, dz;
  CSphere ();
  CSphere (float radius, int segments, float dx, float dy, float dz);
  ~CSphere ();
  void init (float radius, int segments);
  void init (float radius, int segments, float dx, float dy, float dz, int randomized);
  void invertNormals ();
  void setNorthPoleColor (CColor *c, float w);
  void setSouthPoleColor (CColor *c, float w);
  void setPoleColor (int phi, int theta, CColor *c, float w);
};

#endif

