/*
    GL-117
    Copyright 2001, 2002 Thomas A. Drexl aka heptargon

    This file is part of GL-117.

    GL-117 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    GL-117 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Jabytad; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef IS_ULAND_H
#define IS_ULAND_H

//---------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//---------------------------------------------------------------------------

#define MAXX 512
#define MAXX_8 64
#define DIVISIONS 4
#define DIVX 128
#define DIVX_8 16
#define LOG2MAXX 9
#define LOG2SPLINE 9
#define HEIGHT 120

#define GRASS 0
#define CONIFEROUSWOODS0 1
#define CONIFEROUSWOODS1 2
#define CONIFEROUSWOODS2 3
#define CONIFEROUSWOODS3 4
#define DECIDUOUSWOODS0 5
#define DECIDUOUSWOODS1 6
#define DECIDUOUSWOODS2 7
#define DECIDUOUSWOODS3 8
#define MIXEDWOODS0 9
#define MIXEDWOODS1 10
#define MIXEDWOODS2 11
#define MIXEDWOODS3 12
#define DWARFPINES0 13
#define DWARFPINES1 14
#define DWARFPINES2 15
#define DWARFPINES3 16
#define BUSHES0 17
#define BUSHES1 18
#define BUSHES2 19
#define BUSHES3 20
#define ROCKS 30
#define ROCKS2 31
#define GLACIER 32
#define REDSTONE 35
#define REDSAND 36
#define REDTREE0 37
#define WATER 40
#define SHALLOWWATER 41
#define DEEPWATER 42
#define XSHALLOWWATER 43
#define XDEEPWATER 44
#define MOONSAND 50
#define TRYLAKE 100


class Landscape
{
  public:
  unsigned short h [MAXX + 1] [MAXX + 1];
  unsigned short hw [MAXX + 1] [MAXX + 1];
  unsigned char f [MAXX + 1] [MAXX + 1];
  unsigned char ftry [MAXX + 1] [MAXX + 1];
  int type;

  private:
  int highestpoint, lowestpoint, rockborder;
  int maxx;
  int hoehe, maxn, n; // these need not be saved...
  int random (int n);
  int extremerandom (int n);
  int getX (int x);

  public:
  int hg [MAXX + 1] [MAXX + 1];
  void smoothGlacier (); // special erosion function
  void gauss (); // gaussian smooth function for the entire height mask 5x5
  void gaussAlpine (); // gaussian smooth function for the entire height mask, 5x5 low, 3x3 high
  void gauss (int x, int y); // local gaussian smooth function, 5x5
  bool isType (unsigned char type, unsigned char id);
  bool isWoods (int type);
  bool isWater (int type);
  bool isGlacier (int type);
  void init ();
  void genSurface (int hoehepc, int *heightmap);
  void genCanyonSurface (int hoehepc);
  void genMoonSurface ();
  void genRocks (int diffmin, int percent);
//  void cLakeRecursion (short x, short y, unsigned short level, unsigned char sn, int extent);
  int calcLake (int ys, int xs, unsigned short level, int num, int maxextent);
  void genLake (int depthpc);
  void calcWoods (int dy);
  bool isGround (int x, int y);
  bool riverCheck (int x, int y, int *fl, int z, int z2);
  void genRiver ();
  void searchPlain (int divx, int divy, int *x, int *y);
  Landscape ();
  ~Landscape ();
};

#endif

